"""
Base information for pysal meta package
"""


federation_hierarchy = {
    'explore': ['esda', 'giddy', 'segregation',
                'pointpats', 'inequality',
                 'spaghetti'],
    'model': ['spreg', 'spglm', 'tobler', 'spint',
              'mgwr', 'spvcm', 'access'],
    'viz': ['splot', 'mapclassify'],
    'lib': ['libpysal']
}

memberships = {'libpysal': 'lib',
               'esda': 'explore',
               'giddy': 'explore',
               'segregation': 'explore',
               'pointpats': 'explore',
               'inequality': 'explore',
               'spaghetti': 'explore',
               'access': 'model',
               'spreg': 'model',
               'spglm': 'model',
               'spint': 'model',
               'tobler': 'model',
               'spvcm': 'model',
               'splot': 'viz',
               'mapclassify': 'viz'
}


class cached_property(object):
    """ A property that is only computed once per instance and then replaces
        itself with an ordinary attribute. Deleting the attribute resets the
        property.

        Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
        """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value

def _installed_version(package):
    try:
        exec(f'import {package}')
    except ModuleNotFoundError:
        v = 'NA'
    try:
        v = eval(f'{package}.__version__')
    except AttributeError:
        v = 'NA'
    return v

def _installed_versions():
    ver = {}
    for package in memberships.keys():
        ver[package] = _installed_version(package)
    return ver

def _released_versions():
    from .frozen import frozen_packages
    return frozen_packages


class Versions:
    @cached_property
    def installed(self):
        """
        Inventory versions of pysal packages that are installed

        Attributes
        ----------
        installed: dict
                   key is package name, value is version string
        """
        return _installed_versions()

    @cached_property
    def released(self):
        """
        Inventory versions of pysal packages that are released in the meta
        package.

        Attributes
        ----------
        released: dict
                   key is package name, value is version string
        """

        return _released_versions()

    def check(self):
        """
        Print a tabular string that reports installed and released versions of
        PySAL packages.
        """
        table = []
        package = "Package"
        installed = "Installed"
        released = "Released"
        match = "Match"
        s = f'{package:>12} | {installed:>15} | {released:>15} | {match:>5}'
        table.append(s)
        table.append("-"*len(s))
        for package in self.installed:
            installed = self.installed[package]
            released = self.released[package]
            match = installed == released
            s = f'{package:>12} | {installed:>15} | {released:>15} | {match:>5}'
            table.append(s)
        print("\n".join(table))


versions = Versions()




print("import: 'pysal'")
import pysal

print("import: 'pysal.viz'")
import pysal.viz

print("import: 'pysal.explore'")
import pysal.explore

print("import: 'pysal.model'")
import pysal.model

print("import: 'pysal.lib'")
import pysal.lib




from spreg.ols import *
from spreg.diagnostics import *
from spreg.diagnostics_sp import *
from spreg.diagnostics_sur import *
from spreg.diagnostics_tsls import *
from spreg.user_output import *
from spreg.twosls import *
from spreg.twosls_sp import *
from spreg.error_sp import *
from spreg.error_sp_het import *
from spreg.error_sp_hom import *
from spreg.ols_regimes import *
from spreg.twosls_regimes import *
from spreg.twosls_sp_regimes import *
from spreg.error_sp_regimes import *
from spreg.error_sp_het_regimes import *
from spreg.error_sp_hom_regimes import *
from spreg.probit import *
from spreg.ml_lag import *
from spreg.ml_lag_regimes import *
from spreg.ml_error import *
from spreg.ml_error_regimes import *
from spreg.sur import *
from spreg.sur_error import *
from spreg.sur_lag import *
from spreg.sur_utils import *
from spreg.utils import *
from spreg.regimes import *
from spreg.sputils import *



import importlib


def lazy_import(importer_name, to_import):
    """Return the importing module and a callable for lazy importing.

    The module named by importer_name represents the module performing the
    import to help facilitate resolving relative imports.

    to_import is an iterable of the modules to be potentially imported (absolute
    or relative). The `as` form of importing is also supported,
    e.g. `pkg.mod as spam`.

    This function returns a tuple of two items. The first is the importer
    module for easy reference within itself. The second item is a callable to be
    set to `__getattr__`.
    """
    module = importlib.import_module(importer_name)
    import_mapping = {}
    for name in to_import:
        importing, _, binding = name.partition(' as ')
        if not binding:
            _, _, binding = importing.rpartition('.')
        import_mapping[binding] = importing

    def __getattr__(name):
        if name not in import_mapping:
            message = f'module {importer_name!r} has no attribute {name!r}'
            raise AttributeError(message)
        importing = import_mapping[name]
        # imortlib.import_module() implicitly sets submodules on this module as
        # appropriate for direct imports.
        imported = importlib.import_module(importing,
                                           module.__spec__.parent)
        setattr(module, name, imported)
        return imported

    return module, __getattr__
